/*
 *
 *  Copyright (C) 1996-2001, OFFIS
 *
 *  This software and supporting documentation were developed by
 *
 *    Kuratorium OFFIS e.V.
 *    Healthcare Information and Communication Systems
 *    Escherweg 2
 *    D-26121 Oldenburg, Germany
 *
 *  THIS SOFTWARE IS MADE AVAILABLE,  AS IS,  AND OFFIS MAKES NO  WARRANTY
 *  REGARDING  THE  SOFTWARE,  ITS  PERFORMANCE,  ITS  MERCHANTABILITY  OR
 *  FITNESS FOR ANY PARTICULAR USE, FREEDOM FROM ANY COMPUTER DISEASES  OR
 *  ITS CONFORMITY TO ANY SPECIFICATION. THE ENTIRE RISK AS TO QUALITY AND
 *  PERFORMANCE OF THE SOFTWARE IS WITH THE USER.
 *
 *  Module:  wlistctn
 *
 *  Author:  Sandra von Gehlen
 *
 *  Purpose: Modality Worklist Database
 *
 *  Last Update:      $Author: joergr $
 *  Update Date:      $Date: 2001/11/28 13:43:07 $
 *  Source File:      $Source: /share/dicom/cvs-depot/dcmtk/wlistctn/libsrc/wrklstdb.cc,v $
 *  CVS/RCS Revision: $Revision: 1.36 $
 *  Status:           $State: Exp $
 *
 *  CVS/RCS Log at end of file
 *
 */

#include "osconfig.h"

#ifdef HAVE_STDLIB_H
#ifndef  _BCB4
/* workaround for bug in Borland C++ Builder 4 */
BEGIN_EXTERN_C
#endif
#include <stdlib.h>
#ifndef  _BCB4
END_EXTERN_C
#endif
#endif

BEGIN_EXTERN_C
#include <stdio.h>
#include <string.h>
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_LIBC_H
#include <libc.h>
#endif
END_EXTERN_C

#include "ofstring.h"

#include "dctk.h"
#include "dcdebug.h"
#include "dcvm.h"
#include "dctypes.h"
#include "dcstack.h"
#include "dclist.h"
#include "dcvrat.h"
#include "wrklstdb.h"
#include "dcompat.h"
#include "diutil.h"
#include "ofconsol.h"

#define LOCKFILENAME "lockfile"

char * standardizeDate(const char * da)
    /*
     * This function creates a standard string which represents a certain date value
     * and returns this string. If an error occured or an invalid parameter was passed,
     * the NULL pointer will be returned.
     *
     * Parameters:
     *   da - [in] Date value which shall be standardized.
     */
{
    if (!da)
        return NULL;

    if (strlen(da) < 8)
        return NULL;

    char * stdDate = new char[9];
    stdDate[8] = 0;
    int i = 0;
    int error = 0;
    const char * ptr = da;

    while (*ptr && ! error)
    {
        if (*ptr >= '0' && *ptr <='9')
        {
            if (i < 8)
                stdDate[i] = *ptr;
            i++;
        }
        else if ((*ptr != '.' || (i != 4 && i != 6)) && *ptr != ' ')
            error = 1;

        ptr++;
    }

    if (error || i != 8)
    {
        delete[] stdDate;
        return NULL;
    }
    else
        return stdDate;
}

char * standardizeTime(const char *timeString)
    /*
     * This function creates a standard string which represents a certain time value
     * and returns this string. If an error occured or an invalid parameter was passed,
     * the NULL pointer will be returned.
     *
     * Parameters:
     *   timeString - [in] Time value which shall be standardized.
     */
{
    if (!timeString)
        return NULL;

    if (strlen(timeString) < 2)
        return NULL;

    char * stdTime = new char[14];
    stdTime[13]=0;

    int i = 0;
    int error = 0;
    const char * ptr = timeString;

    while(*ptr && !error)
    {
        if ((*ptr >= '0' && *ptr <='9') ||
            (*ptr == '.' && i == 6))
        {
            if (i < 13)
                stdTime[i] = *ptr;
            i++;
        }
        else if ((*ptr != ':' || (i != 2 && i != 4)) && *ptr != ' ')
            error = 1;
        ptr++;
    }

    if ( i < 2 || ((i & 1) && i<6 ))
        error = 1;

    if (!error)
    {
        if (i < 6)
            for (;i < 6; i++)
                stdTime[i] = '0';

        if (i==6)
            stdTime[i++] = '.';

        if (i < 13)
            for(; i < 13; i++)
                stdTime[i] = '0';

    }

    if (error)
    {
        delete[] stdTime;
        return NULL;
    }
    else
        return stdTime;
}

int WorklistDB::set_readlock(const char *path)
    /*
     * This function sets a read lock on the LOCKFILE in the given directory.
     * The function will return 1 if read lock has been set successfully, else it will return 0.
     *
     * Parameters:
     *   path - [in] The directory path that refers to the directory which contains the worklist files.
     */
{
#ifndef _WIN32
  struct flock lockdata;
#endif
  OFString lockname;
  int result;

  /* if no path is specified, return */
  if (path == NULL)
  {
    CERR << "error: NULL path not allowed" << endl;
    return 0;
  }

  /* if a read lock has already been set, return */
  if (fd != 0)
  {
    CERR << "error: nested read locks not allowed!" << endl;
    return 0;
  }

  /* assign path to a local variable */
  lockname = path;

  /* if the given path does not show a PATH_SEPERATOR at the end, append one */
  if ((lockname.length()>0) && (lockname[lockname.length()-1] != PATH_SEPARATOR))
  {
    lockname += PATH_SEPARATOR;
  }

  /* append LOCKFILENAME to the given path (and seperator) */
  lockname += LOCKFILENAME;

  /* open corresponding file */
  fd=open(lockname.c_str(), O_RDWR);
  if (fd == -1)
  {
    fd=0;
    CERR << "error: cannot open file " << lockname.c_str() << endl
        << "return code: " << strerror(errno) << endl;
    return 0;
  }

  /* now set a read lock on the corresponding file */

#ifdef _WIN32
  /* windows does not have fcntl locking */
  result = dcmtk_flock(fd, LOCK_SH);
#else
  lockdata.l_type = F_RDLCK;
  lockdata.l_whence=0;
  lockdata.l_start=0;
  lockdata.l_len=0;
#if SIZEOF_VOID_P == SIZEOF_INT
  /* some systems, e.g. NeXTStep, need the third argument
   * for fcntl calls to be casted to int. Other systems,
   * e.g. OSF1-Alpha, won't accept this because int and struct flock *
   * have different sizes. The workaround used here is to use a typecast to int
   * if sizeof(void *) == sizeof(int) and leave it away otherwise.
   */
  result = fcntl(fd, F_SETLKW, (int)(&lockdata));
#else
  result = fcntl(fd, F_SETLKW, &lockdata);
#endif
#endif
  if (result == -1)
  {
    CERR << "error: cannot set read lock on file " << lockname.c_str() << endl;
    dcmtk_plockerr("return code");
    close(fd);
    fd=0;
    return 0;
  }

  /* return success */
  return 1;
}

int WorklistDB::release_readlock()
{
#ifndef _WIN32
  struct flock lockdata;
#endif
  int result;

  if (fd==0)
  {
    CERR << "error: no readlock to release" << endl;
    return 0;
  }

#ifdef _WIN32
  /* windows does not have fcntl locking */
  result = dcmtk_flock(fd, LOCK_UN);
#else
  lockdata.l_type = F_UNLCK;
  lockdata.l_whence=0;
  lockdata.l_start=0;
  lockdata.l_len=0;
#if SIZEOF_VOID_P == SIZEOF_INT
  result = fcntl(fd, F_SETLKW, (int)(&lockdata));
#else
  result = fcntl(fd, F_SETLKW, &lockdata);
#endif
#endif
  if (result == -1)
  {
    CERR << "error: cannot release read lock" << endl;
    dcmtk_plockerr("return code");
    close(fd);
    fd=0;
    return 0;
  }

  close(fd);
  fd=0;
  return 1;
}


static
OFBool wcmatch(const char * string, const char * wc);

static
OFBool matchstar(const char * string, const char * wc)
    /*
     * This function is called, if the search pattern contains a star symbol. It determines if the string
     * value which was passed as argument 1 still matches the pattern value which was passed as argument 2.
     * This function takes the star symbol in wc into account. (Note that the pattern value might contain
     * more wild card symbols.) The function will return OFTrue if there is a match; if there is not a match
     * it will return OFFalse.
     *
     * Parameters:
     *   string - [in] Value which shall be matched against pattern.
     *   wc     - [in] Matching pattern (may contain wild card symbols).
     */
{
    /* initialize result value */
    OFBool ok = OFFalse;

    /* assign pointers which were passed to local variables */
    const char * strptr = string;
    const char * wcptr = wc;

    /* move pointer one char to the right as long as it points to a star symbol */
    while (*wcptr == '*')
        wcptr++;

    /* if we got to the end of the matching pattern, return OFTrue */
    if (*wcptr == 0)
        return OFTrue;

    /* if there is something else at the end of the matching pattern, */
    /* we need to go ahead and compare the rest of the two strings */

    /* as long as ok equals OFFalse and we did not get to the end of the string */
    while (!ok && *strptr != 0)
    {
        /* if the pattern reveals a '?' or if both pointers refer to the same */
        /* character, we need to call wcmatch again, to determine a result */
        if (*wcptr == '?' || *strptr == *wcptr)
            ok = wcmatch(strptr+1, wcptr+1);

        /* if ok still equals OFFalse, set pointer one character to the right */
        if (!ok)
            strptr++;
    }

    /* return result */
    return ok;
}

static
OFBool wcmatch(const char * string, const char * wc)
    /*
     * This function determines if the string value which was passed as argument 1 matches the pattern value
     * which was passed as argument 2. (Note that the pattern value might contain wild card symbols.) The
     * function will return OFTrue if this is the case; if this is not the case it will return OFFalse.
     *
     * Parameters:
     *   string - [in] Value which shall be matched against pattern.
     *   wc     - [in] Matching pattern (may contain wild card symbols).
     */
{
    /* if there is no matching pattern return false */
    if (wc == NULL)
        return OFFalse;

    /* generate a new pointer to a string from origin string */
    const char * strptr;
    if (string == NULL)
        strptr = "";
    else
        strptr = string;

    /* generate a new pointer to a pattern string from origin pattern */
    const char * wcptr = wc;

    /* initialize return value */
    OFBool ok = OFFalse;

    /* go through both strings character by character */
    while(*wcptr != 0 && *wcptr != '*' &&
          ((*wcptr == '?' && *strptr != 0) ||  *wcptr == *strptr))
    {
        wcptr++;
        strptr++;
    }

    /* if the current pattern character equals the star symbol (wild card symbol) we need */
    /* to call another function. If this is not the case, ok will be set to OFTrue or */
    /* OFFalse, depending on if we got to the end of both strings in the above loop. */
    if (*wcptr == '*')
        ok = matchstar(strptr, wcptr+1);
    else
        ok = *wcptr == 0 && *strptr == 0;

    /* return result */
    return ok;
}

static char* new_strdup(const char* s)
{
    if (s == NULL) return NULL;
    int len = strlen(s);
    char* ss = new char[len+1];
    strcpy(ss, s);
    return ss;
}

static OFBool matchPattern(const char* value, const char* pattern)
    /*
     * This function determines if the value which was passed matches the pattern which was passed.
     * The function will return OFTrue if this is the case; if this is not the case it will return OFFalse.
     *
     * Parameters:
     *   value   - [in] Value which shall be matched against pattern.
     *   pattern - [in] Matching pattern.
     */
{
    /* intialize return value */
    OFBool ok = OFFalse;

    /* make copies and strip trailing spaces */
    value = DU_stripTrailingSpaces(new_strdup(value));
    pattern = DU_stripTrailingSpaces(new_strdup(pattern));

    /* if there is no pattern or if the pattern is an empty string, it is considered to be a match */
    if (pattern == NULL || pattern[0] == '\0')
    {
        ok = OFTrue;
    }
    /* else if the string value matches the pattern (also taking into account */
    /* wild card symbols), it is for sure considered to be a match */
    else if (wcmatch(value, pattern))
    {
        ok = OFTrue;
    }

    /* free memory */
    delete[] (char*)value;
    delete[] (char*)pattern;

    /* return result */
    return ok;
}

static OFBool
rangematchTime (const char * s1, const char * range)
{
    if (s1 == NULL || range == NULL)
        return OFFalse;

    char * t;
    OFBool ok = OFFalse;
    int len = strlen(range);
    t = new char[len+1];
    strcpy(t, range);
    char * strich = strchr(t, '-');

    if (strchr(t, '-'))
    {
        if (t[0] == '-')
        {
            char *n1;
            n1 = standardizeTime(&t[1]);
            if (n1 != NULL)
            {
                if (strcmp(s1, n1) <= 0)
                    ok = OFTrue;
                else
                    ok = OFFalse;
            }
            else
                ok = OFFalse;
            delete[] n1;
        }
        else if ((t[len-1] == '-') && (t[len] == '\0'))
        {
            char *n1;
            t[len-1] ='\0';
            n1 =  standardizeTime(t);
            if (n1 != NULL)
            {
                if (strcmp(s1, n1) >= 0)
                    ok = OFTrue;
                else
                    ok = OFFalse;
            }
            else
                ok = OFFalse;
            delete[] n1;
        }
        else
        {
           * strich = 0;
           strich ++;
           char *n2;
           char *n1;
           n1 = standardizeTime(t);
           n2 = standardizeTime(strich);
           if ((n1 != NULL)&&(n1 != NULL))
           {
               if ((strcmp(s1, n1) >= 0) && (strcmp(s1, n2) <= 0))
                   ok = OFTrue;
               else
                   ok = OFFalse;
           }
           else
               ok = OFFalse;
           delete[] n1;
           delete[] n2;
        }
    }
    else
        ok = OFFalse;
    delete[] t;
    return ok;
}

static OFBool
matchTime(const char* value, const char* pattern)
    /*
     * This function determines if the (time) value which was passed as a first argument matches
     * the (time) pattern which was passed as a second argument. The function will return OFTrue
     * if this is the case; if this is not the case it will return OFFalse.
     *
     * Parameters:
     *   value   - [in] Value which shall be matched against pattern.
     *   pattern - [in] Matching pattern.
     */
{
    /* intialize result variable */
    OFBool ok = OFFalse;

    /* make copies and strip trailing spaces */
    value = DU_stripTrailingSpaces(new_strdup(value));
    pattern = DU_stripTrailingSpaces(new_strdup(pattern));

    /* if there is no or only an empty pattern, this is considered to be a match */
    if (pattern == NULL || pattern[0] == '\0')
    {
        ok = OFTrue;
    }
    /* else if the value which shall be compared to pattern is NULL or empty, */
    /* this is considered to be NO match since the pattern has contents */
    else if (value == NULL || value[0] == '\0')
    {
        ok = OFFalse;
    }
    /* in all other cases do the following */
    else
    {
        /* create a standard string which represents the */
        /* value which shall be compared to the pattern */
        char *stdValue;
        stdValue = standardizeTime(value);

        /* if a '-' character cannot be found in the pattern */
        if (!strchr(pattern,'-'))
        {
            /* create a standard string which represents the pattern */
            char *stdPattern;
            stdPattern = standardizeTime(pattern);

            /* if the standardized pattern is valid and it equals the standardized */
            /* value, ok will be set to OFTrue, else ok will be set to OFFalse */
            if (stdPattern != NULL)
            {
                ok = (strcmp(stdValue, stdPattern) == 0);
            }
            else
            {
                ok = OFFalse;
            }

            /* free memory */
            delete[] stdPattern;
        }
        /* else if a '-' character CAN be found in the pattern, we need */
        /* to call another function to determine the comparison's result */
        else
        {
            ok = rangematchTime(stdValue, pattern);
        }

        /* free memory */
        delete[] stdValue;
    }

    /* free memory */
    delete[] (char*)value;
    delete[] (char*)pattern;

    /* return result */
    return ok;
}

static OFBool
rangematchDate (const char * s1, const char * range)
{
    if (s1 == NULL || range == NULL)
        return OFFalse;
    char * t;
    OFBool ok = OFFalse;

    int len = strlen(range);
    t = new char[len+1];
    strcpy(t, range);
    char * strich = strchr(t, '-');

    if (strchr(t, '-'))
    {
        if (t[0] == '-')
        {
            char *n1;
            n1 = standardizeDate(&t[1]);
            if (n1 != NULL)
            {
                if (strcmp(s1, n1) <= 0)
                    ok = OFTrue;
                else
                    ok = OFFalse;
            }
            else
                ok = OFFalse;
            delete[] n1;
        }
        else if ((t[len-1] == '-') && (t[len] == '\0'))
        {
            char *n1;
            t[len-1] ='\0';
            n1 =  standardizeDate(t);
            if (n1 != NULL)
            {
                if (strcmp(s1, n1) >= 0)
                    ok = OFTrue;
                else
                    ok = OFFalse;
            }
            else
                ok = OFFalse;
            delete[] n1;
        }
        else
        {
           * strich = 0;
           strich ++;
           char *n2;
           char *n1;
           n1 = standardizeDate(t);
           n2 = standardizeDate(strich);
           if ((n1 != NULL)&&(n2 != NULL))
           {
               if ((strcmp(s1, n1) >= 0) && (strcmp(s1, n2) <= 0))
                   ok = OFTrue;
               else
                   ok = OFFalse;
           }
           else
               ok = OFFalse;
           delete[] n1;
           delete[] n2;
        }
    }
    else
        ok = OFFalse;
    delete[] t;
    return ok;
}

static OFBool
matchDate(const char* value, const char* pattern)
    /*
     * This function determines if the (date) value which was passed as a first argument matches
     * the (date) pattern which was passed as a second argument. The function will return OFTrue
     * if this is the case; if this is not the case it will return OFFalse.
     *
     * Parameters:
     *   value   - [in] Value which shall be matched against pattern.
     *   pattern - [in] Matching pattern.
     */
{
    /* intialize result variable */
    OFBool ok = OFFalse;

    /* make copies and strip trailing spaces */
    value = DU_stripTrailingSpaces(new_strdup(value));
    pattern = DU_stripTrailingSpaces(new_strdup(pattern));

    /* if there is no or only an empty pattern, this is considered to be a match */
    if (pattern == NULL || pattern[0] == '\0')
    {
        ok = OFTrue;
    }
    /* else if the value which shall be compared to pattern is NULL or empty, */
    /* this is considered to be NO match since the pattern has contents */
    else if (value == NULL || value[0] == '\0')
    {
        ok = OFFalse;
    }
    /* in all other cases do the following */
    else
    {
        /* create a standard string which represents the */
        /* value which shall be compared to the pattern */
        char *stdValue;
        stdValue = standardizeDate(value);

        /* if a '-' character cannot be found in the pattern */
        if (!strchr(pattern,'-'))
        {
            /* create a standard string which represents the pattern */
            char *stdPattern;
            stdPattern = standardizeDate(pattern);

            /* if the standardized pattern is valid and it equals the standardized */
            /* value, ok will be set to OFTrue, else ok will be set to OFFalse */
            if (stdPattern != NULL)
            {
                ok = (strcmp(stdValue, stdPattern) == 0);
            }
            else
            {
                ok = OFFalse;
            }

            /* free memory */
            delete[] stdPattern;
        }
        /* else if a '-' character CAN be found in the pattern, we need */
        /* to call another function to determine the comparison's result */
        else
        {
            ok = rangematchDate(stdValue, pattern);
        }

        /* free memory */
        delete[] stdValue;
    }

    /* free memory */
    delete[] (char*)value;
    delete[] (char*)pattern;

    /* return result */
    return ok;
}

/**
 * Match combined date and times.
 * the currently implemented range matching algorithms do not properly
 * handle this problem.
 * If ranges are fully specified for both date and time
 * (e.g. date="19971101-19971102"
 * and time="0900-1800") then this should match the outer product (e.g. the
 * period between "1 Nov. 1997, 9am" until "2 Nov. 1997, 6pm").
 * If a range is fully specified for only one of date and time then the
 * inner product is intended (e.g. date="19971105-19971106" and time="1100-"
 * would match "5 Nov. 1997, 2pm" but _not_ "6 Nov. 1997, 9am".
 *
 * ***** NOT YET COMPLETE *****
 *
 * @param date the date to match (always matches if NULL).
 * @param time the time to match (always matches if NULL).
 * @param dateRange the date range to match against (always matches
 * if NULL).
 * @param timeRange the time range to match against (always matches
 * if NULL).
 * @return true if the date/time combination matches the ranges.
 */

#if 0  /* ***** NOT YET COMPLETE ***** */
static OFBool
matchDateAndTime(const char* date, const char* time,
                 const char* dateRange, const char* timeRange)
{
    OFBool ok = OFFalse;
    if (timeRange == NULL) {
        ok = matchDate(date, dateRange);
    } else if (dateRange == NULL) {
        ok = matchTime(time, timeRange);
    } else if (!(strchr(dateRange, '-') && strchr(timeRange, '-'))) {
        // both are not proper ranges
        ok = (matchDate(date, dateRange) && matchTime(time, timeRange));
    } else {
        // both are ranges but the date must match anyway
        ok = matchDate(date, dateRange);
        // fine tune by time

        /*
        ** Recheck all matching routines for padding problems.
        ** What happens with a range "11- " ???
        **
        **/
        // ***** NOT YET COMPLETE *****

    }
    return ok;
}
#endif  /* ***** NOT YET COMPLETE ***** */

static OFBool
isaSupportedMatchingKey(const DcmTagKey& key, int level)
    /*
     * This function checks if the given tag refers to an attribute which is a supported
     * matching attribute. Note that there are 2 levels of supported matching attributes:
     * - Level 0 supports the attributes DCM_PatientsName, DCM_PatientID, and DCM_ScheduledProcedureStepSequence.
     * - Level 1 supports the attributes DCM_ScheduledStationAETitle, DCM_ScheduledProcedureStepStartDate,
     *   DCM_ScheduledProcedureStepStartTime, DCM_Modality, DCM_ScheduledPerformingPhysiciansName,
     *   DCM_ScheduledStationName, DCM_ScheduledProcedureStepLocation, and DCM_CommentsOnTheScheduledProcedureStep.
     *
     * Parameters:
     *   key   - [in] Contains the tag which will be checked.
     *   level - [in] Specifies on which level the checking shall be conducted.
     */
{
    /* initialize result variable */
    OFBool ok = OFFalse;

    /* depending on the level, conduct the checking */
    if (level == 0) {
        ok = (key == DCM_PatientsName) ||
            (key == DCM_PatientID) ||
            (key == DCM_ScheduledProcedureStepSequence);
    } else if (level == 1) {
        ok = (key == DCM_ScheduledStationAETitle) ||
            (key == DCM_ScheduledProcedureStepStartDate) ||
            (key == DCM_ScheduledProcedureStepStartTime) ||
            (key == DCM_Modality) ||
            (key == DCM_ScheduledPerformingPhysiciansName) ||
            (key == DCM_ScheduledStationName) ||
            (key == DCM_ScheduledProcedureStepLocation) ||
            (key == DCM_CommentsOnTheScheduledProcedureStep);
    }

    /* return result */
    return ok;
}

static OFBool
isaSupportedReturnKey(const DcmTagKey& key, int level)
    /*
     * This function checks if the given tag refers to an attribute which is a supported
     * return attribute. Note that there are 2 levels of supported return attributes:
     * - Level 0 supports the attributes DCM_PatientsName, DCM_PatientID, and DCM_ScheduledProcedureStepSequence
     *   (accounted for in isaSupportedMatchingKey(...)) plus the following attributes:
     *   DCM_SpecificCharacterSet, DCM_MedicalAlerts, DCM_ContrastAllergies, DCM_RequestedProcedureID
     *   DCM_RequestedProcedureDescription, DCM_StudyInstanceUID, DCM_RequestedProcedurePriority,
     *   DCM_AccessionNumber, DCM_RequestingPhysician, DCM_PatientsBirthDate, DCM_PatientsSex,
     *   DCM_PatientTransportArrangements, DCM_ReferringPhysiciansName, DCM_AdmissionID,
     *   DCM_CurrentPatientLocation, DCM_PatientsWeight, DCM_ConfidentialityConstraintOnPatientDataDescription,
     *   DCM_PatientState, DCM_PregnancyStatus, and DCM_SpecialNeeds.
     * - Level 1 supports the attributes DCM_ScheduledStationAETitle, DCM_ScheduledProcedureStepStartDate,
     *   DCM_ScheduledProcedureStepStartTime, DCM_Modality, DCM_ScheduledPerformingPhysiciansName,
     *   DCM_ScheduledStationName, DCM_ScheduledProcedureStepLocation, and DCM_CommentsOnTheScheduledProcedureStep
     *   (accounted for in isaSupportedMatchingKey(...)) plus the following attributes:
     *   DCM_ScheduledProcedureStepDescription, DCM_PreMedication, DCM_ScheduledProcedureStepID, and
     *   DCM_RequestedContrastAgent.
     *
     * Parameters:
     *   key   - [in] Contains the tag which will be checked.
     *   level - [in] Specifies on which level the checking shall be conducted.
     */
{
    /* initialize result variable */
    OFBool ok = OFFalse;

    /* depending on the level, conduct the checking */
    if (level == 0) {
        ok = isaSupportedMatchingKey(key, level) ||
            (key == DCM_SpecificCharacterSet) ||
            (key == DCM_MedicalAlerts) ||
            (key == DCM_ContrastAllergies) ||
            (key == DCM_RequestedProcedureID) ||
            (key == DCM_RequestedProcedureDescription) ||
            (key == DCM_StudyInstanceUID) ||
            (key == DCM_RequestedProcedurePriority) ||
            (key == DCM_AccessionNumber) ||
            (key == DCM_RequestingPhysician) ||
            (key == DCM_PatientsBirthDate) ||
            (key == DCM_PatientsSex) ||
            (key == DCM_PatientTransportArrangements) ||
            (key == DCM_ReferringPhysiciansName) ||
            (key == DCM_AdmissionID) ||
            (key == DCM_CurrentPatientLocation) ||
            (key == DCM_PatientsWeight) ||
            (key == DCM_ConfidentialityConstraintOnPatientDataDescription) ||
            (key == DCM_PatientState) ||
            (key == DCM_PregnancyStatus) ||
            (key == DCM_SpecialNeeds);
   } else if (level == 1) {
        ok = isaSupportedMatchingKey(key, level) ||
            (key == DCM_ScheduledProcedureStepDescription) ||
            (key == DCM_PreMedication) ||
            (key == DCM_ScheduledProcedureStepID) ||
            (key == DCM_RequestedContrastAgent);
    }

    /* return result */
    return ok;
}

static char *
getStringValue(DcmElement* elem)
    /*
     * This function returns the value of the given DICOM string element (attribute).
     * If the element does not refer to a string attribute, a NULL pointer is returned.
     *
     * Parameters:
     *   elem - [in] The DICOM element.
     */
{
    char *val = NULL;
    if ((elem) && (elem->isaString())) elem->getString(val);
    return val;
}


static char *
findStringValue(DcmItem* dset, const DcmTagKey &tag)
    /*
     * This function looks for a DICOM string element with tag 'tag' in the data set which
     * was passed as first argument and returns the value of this element (attribute).
     * If the element does not refer to a string attribute, a NULL pointer is returned.
     *
     * Parameters:
     *   dset - [in] Data set which contains DICOM elements.
     *   tag  - [in] Element (attribute) that shall be looked for in dset.
     */
{
    DcmStack stk;
    char *val=NULL;
    dset->search(tag, stk, ESM_fromHere, OFFalse);
    DcmObject *obj = stk.top();
    if (obj) val = getStringValue((DcmElement*)obj);
    return val;
}

static Uint16 *
findUSValue(DcmItem* dset, const DcmTagKey &tag)
{
    static Uint16 val;
    DcmStack stk;

    Uint16 *result=NULL;
    dset->search(tag, stk, ESM_fromHere, OFFalse);
    DcmObject *obj = stk.top();
    if (obj)
    {
      ((DcmElement*)obj)->getUint16(val,0);
      result = &val;
    }
    return result;
}

// *******************************


WorklistDB::WorklistDB(const char* dbDirectoryPath)
: dirpath(dbDirectoryPath)
#ifdef HAVE__FINDFIRST
#else
, dirp(NULL)
#endif
, identifiers(NULL)
, objlist(NULL)
, errorElements(NULL)
, offendingElements(NULL)
, errorComment(NULL)
, fd(0)
, foundUnsupportedOptionalKey(OFFalse)
, failInvalidQuery(OFTrue)
, verbose(OFFalse)
{
    // make sure data dictionary is loaded
    if (!dcmDataDict.isDictionaryLoaded()) {
        CERR << "Warning: no data dictionary loaded, check environment variabe: "
             << DCM_DICT_ENVIRONMENT_VARIABLE << endl;
    }
    objlist = new DcmList();
    identifiers = new DcmDataset();
    offendingElements = new DcmAttributeTag(DCM_OffendingElement, 0);
    errorElements = new DcmAttributeTag(DCM_OffendingElement, 0);
    errorComment = new DcmLongString(DCM_ErrorComment, 0);
}

// *******************************

WorklistDB::~WorklistDB()
{
    this -> clearObjectList();
    delete objlist;
    this -> clearDataset(identifiers);
    delete identifiers;
    delete offendingElements;
    delete errorElements;
    delete errorComment;
    if (fd != 0) release_readlock();
}

// *******************************

void WorklistDB::clearObjectList(void)
    /*
     * This function removes all elements from member variable objlist. This variable
     * will be used to capture all records that match the specified search mask.
     *
     * Parameters:
     *   none.
     */
{
    /* if the list is defined and there are elements in the list */
    if (objlist && objlist -> card())
    {
        /* go through the list of elements, starting at the first element */
        DcmObject * removeObj = NULL;
        objlist -> seek(ELP_first);
        for(unsigned long i = 0; i < objlist->card(); i++)
        {
            /* and remove all elements */
            removeObj = objlist->remove();
            if (removeObj)
                delete removeObj;
        }
    }
}

// *******************************

void WorklistDB::clearDataset(DcmDataset* idents)
{
    if ( idents && (idents->card() > 0))
    {
        idents->clear();
    }
}

// *******************************

void WorklistDB::putOffendingElements(DcmTagKey& tag)
{
    DcmTagKey errortag;
    unsigned long d = offendingElements->getVM();

    if ( d == 0)
    {
        offendingElements->putTagVal(tag, 0);
    }
    else
    {
        for (unsigned long j = 0; j < d; j++)
        {
            offendingElements->getTagVal(errortag, j);
            if (errortag != tag)
            {
                offendingElements->putTagVal(tag, d);
            }
        }
    }
}

// *******************************

void WorklistDB::putErrorElements(DcmTagKey & tag)
{
    errorElements->putTagVal(tag, errorElements->getVM());
}

//*****************************

static OFBool
checkCharSet(const char* s, const char* charset)
{
    if (s == NULL || charset == NULL) {
        return OFTrue; /* an emtpy string matches everything */
    }
    /*
    ** return true if all the characters of s can be found
    ** in the string charset.
    */
    int s_len = strlen(s);
    int charset_len = strlen(charset);
    for (int i=0; i<s_len; i++) {
        OFBool isSetMember = OFFalse;
        for (int j=0; !isSetMember && j<charset_len; j++) {
            isSetMember = (s[i] == charset[j]);
        }
        if (!isSetMember) {
            return OFFalse;
        }
    }
    return OFTrue;
}

//*****************************

OFBool WorklistDB::checkMatchingKey(DcmElement *elem)
    /*
     * This function checks if the passed element's value only uses characters
     * which are part of the element's data type's character repertoire.
     *
     * In general, note that the checking needs to be more comprehensive here.
     * The checks here only cover a few of the problems which can be encountered.
     *
     * Parameters:
     *   elem - [in] Element which shall be checked.
     */
{
    OFBool ok = OFTrue;
    char *val = NULL;

    switch (elem->ident()) {
    case EVR_DA:
        val = getStringValue(elem);
        /* although normally DA is 8 bytes fixed length, in a query it
        ** can be a date range (with a range matching symbol '-') and
        ** thus need space padding (max 18 bytes).
        */
        if (!checkCharSet(val, "0123456789.- ")) {
            DcmTag tag(elem->getTag());
            this->putOffendingElements(tag);
            errorComment->putString("Invalid Character Repertoir for DA");
            ok = OFFalse;
        }
        break;

    case EVR_TM:
        val = getStringValue(elem);
        /* although normally TM is max 16 bytes length, in a query it
        ** can be up to 33 bytes (with a range matching symbol '-') and
        ** thus need space padding.
        */
        if (!checkCharSet(val, "0123456789.:- ")) {
            DcmTag tag(elem->getTag());
            this->putOffendingElements(tag);
            errorComment->putString("Invalid Character Repertoir for TM");
            ok = OFFalse;
        }
        break;

    case EVR_CS:
        val = getStringValue(elem);
        /* possible wildcards & space padding */
        if (!checkCharSet(val, "*.ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789 _")) {
            DcmTag tag(elem->getTag());
            this->putOffendingElements(tag);
            errorComment->putString("Invalid Character Repertoir for CS");
            ok = OFFalse;
        }
        break;

    case EVR_AS:
        val = getStringValue(elem);
        /* possible wildcards & space padding */
        if (!checkCharSet(val, "*.0123456789DWMY ")) {
            DcmTag tag(elem->getTag());
            this->putOffendingElements(tag);
            errorComment->putString("Invalid Character Repertoir for AS");
            ok = OFFalse;
        }
        break;

    case EVR_DS:
        val = getStringValue(elem);
        /* possible space padding */
        if (!checkCharSet(val, "0123456789+-Ee. ")) {
            DcmTag tag(elem->getTag());
            this->putOffendingElements(tag);
            errorComment->putString("Invalid Character Repertoir for DS");
            ok = OFFalse;
        }
        break;

    case EVR_IS:
        val = getStringValue(elem);
        /* possible space padding */
        if (!checkCharSet(val, "0123456789+-. ")) {
            DcmTag tag(elem->getTag());
            this->putOffendingElements(tag);
            errorComment->putString("Invalid Character Repertoir for IS");
            ok = OFFalse;
        }
        break;

    case EVR_UI:
        val = getStringValue(elem);
        /* possible multiple values (spaces are not allowed) */
        if (!checkCharSet(val, "0123456789.\\")) {
            DcmTag tag(elem->getTag());
            this->putOffendingElements(tag);
            errorComment->putString("Invalid Character Repertoir for UI");
            ok = OFFalse;
        }
        break;

    default:
        break;
    }

    return ok;
}

//*****************************


OFBool WorklistDB::checkIdentifiers(DcmDataset* idents)
    /*
     * This function checks if all elements in the search mask are supported matching
     * attributes. It returns OFTrue if this is the case, OFFalse if this is not the case.
     *
     * Parameters:
     *   idents - [in] Contains the search mask.
     */
{
    /* initialize counter that counts invalid elements in the search mask */
    int invalidMatchingKeyCount = 0;

    /* remember the number of data elements in the search mask */
    unsigned long cids = idents->card();

    /* dump some information if required */
    if (verbose) {
        COUT << "Checking C-FIND-RQ Identifiers" << endl;
    }

    /* this member variable indicates if we encountered an unsupported */
    /* optional attribute in the search mask; initialize it with false. */
    foundUnsupportedOptionalKey = OFFalse;

    /* start a loop; go through all data elements in the search mask, for each element, do some checking */
    for (unsigned long i = 0; i < cids ; i++)
    {
        /* determine current element */
        DcmElement *element = idents->getElement(i);

        /* determine current element's length */
        Uint32 length = element->getLength();

        /* determine current element's tag */
        DcmTag tag(element->getTag());

        /* if the current element is NOT a sequence */
        if (element->ident() != EVR_SQ)
        {
            /* figure out if the current element refers to an attribute which */
            /* is a supported matching attribute. Use level 0 for this check. */
            if (isaSupportedMatchingKey(tag.getXTag(), 0))
            {
                /* if the current element is a supported matching attribute, do nothing (everything is ok) */
            }
            /* if current element refers to an attribute which is not a supported */
            /* matching attribute, figure out if the current element refers to an */
            /* attribute which is a supported return attribute. Use level 0 for this check. */
            else if (isaSupportedReturnKey(tag.getXTag(), 0))
            {
                /* if this is the case and the current element's length does */
                /* not equal 0, we've found an unsupported optional attribute */
                if (length != 0)
                {
                    foundUnsupportedOptionalKey = OFTrue;
                }
            }
            /* if current element does neither refer to an unsupported matching */
            /* attribute nor does it refer to an unsupported return attribute, */
            /* we've found an unsupported optional attribute; we want to remember */
            /* this attribute in the list of error elements. */
            else
            {
                foundUnsupportedOptionalKey = OFTrue;
                this->putErrorElements(tag);
            }

            /* check if the current element meets certain conditions (at the moment */
            /* we only check if the element's data type and value go together) */
            if (!checkMatchingKey(element))
            {
                /* if there is a problem with the current element, increase the corresponding counter */
                invalidMatchingKeyCount++;
            }
        }
        /* or if the current element IS a sequence */
        else
        {
            /* figure out if the current element refers to an attribute which */
            /* is a supported matching attribute. Use level 0 for this check. */
            if (isaSupportedMatchingKey(tag.getXTag(), 0))
            {
                /* if the current element is a supported matching attribute, check its length */
                if (length != 0)
                {
                    /* if length does not equal 0, check inside the sequence */

                    /* remember that elements is a sequence of items */
                    DcmSequenceOfItems *sequits = (DcmSequenceOfItems*)element;

                    /* check the cardinality of the sequence; note that we may */
                    /* not have more than 1 item in the request identifiers */
                    if (sequits->card() > 1)
                    {
                        this->putOffendingElements(tag);
                        errorComment->putString("More than 1 item in sequence");
                        COUT << "ERROR: More than 1 item in sequence: " << tag.getXTag() << " " << tag.getTagName() << endl;
                        invalidMatchingKeyCount++;
                    }

                    /* get the first item */
                    DcmItem *item = sequits -> getItem(0);

                    /* determine the cardinality of this item */
                    unsigned long cit = item -> card();

                    /* go through all elements of this item */
                    for (unsigned long k = 0; k < cit; k++)
                    {
                        /* determine the current element */
                        DcmElement *elementseq = item -> getElement(k);

                        /* determine the length of the current element */
                        Uint32 length1 = elementseq -> getLength();

                        /* determine the current element's tag */
                        DcmTag tagseq(elementseq->getTag());

                        /* figure out if the current element refers to an attribute which */
                        /* is a supported matching attribute. Use level 1 for this check. */
                        if (isaSupportedMatchingKey(tagseq.getXTag(), 1))
                        {
                            /* if the current element is a supported matching attribute, do nothing (everything is ok) */
                        }
                        /* if current element refers to an attribute which is not a supported */
                        /* matching attribute, figure out if the current element refers to an */
                        /* attribute which is a supported return attribute. Use level 1 for this check. */
                        else if (isaSupportedReturnKey(tagseq.getXTag(), 1))
                        {
                            /* if this is the case and the current element's length does */
                            /* not equal 0, we've found an unsupported optional attribute */
                            if (length1 != 0)
                            {
                                foundUnsupportedOptionalKey = OFTrue;
                            }
                        }
                        /* if current element does neither refer to an unsupported matching */
                        /* attribute nor does it refer to an unsupported return attribute, */
                        /* we've found an unsupported optional attribute; we want to remember */
                        /* this attribute in the list of error elements. */
                        else
                        {
                            foundUnsupportedOptionalKey = OFTrue;
                            this->putErrorElements(tagseq);
                        }

                        /* check if the current element meets certain conditions (at the moment */
                        /* we only check if the element's data type and value go together) */
                        if (!checkMatchingKey(elementseq))
                        {
                            /* if there is a problem with the current element, increase the corresponding counter */
                            invalidMatchingKeyCount++;
                        }
                    }

#ifdef DELETE_UNEXPECTED_ELEMENTS_FROM_IDENTIFIER
                    /* remove unexpected (error) elements from this item */
                    unsigned long d = errorElements->getVM();
                    for (unsigned long e = 0; e < d; e++)
                    {
                        DcmTag errortag;
                        errortag = errorElements->get(e);
                        delete item -> remove(errortag);
                    }
#endif
                }
            }
            /* if current element refers to an attribute which is not a supported */
            /* matching attribute, figure out if the current element refers to an */
            /* attribute which is a supported return attribute. Use level 0 for this check. */
            else if (isaSupportedReturnKey(tag.getXTag(), 0))
            {
                /* if this is the case and the current element's length does */
                /* not equal 0, we've found an unsupported optional attribute */
                if (length != 0)
                {
                    foundUnsupportedOptionalKey = OFTrue;
                }
            }
            /* if current element does neither refer to an unsupported matching */
            /* attribute nor does it refer to an unsupported return attribute, */
            /* we've found an unsupported optional attribute; we want to remember */
            /* this attribute in the list of error elements. */
            else
            {
                foundUnsupportedOptionalKey = OFTrue;
                this->putErrorElements(tag);
            }
        }
    }

#ifdef DELETE_UNEXPECTED_ELEMENTS_FROM_IDENTIFIER
    /* remove unexpected (error) elements from this item */
    unsigned long d = errorElements->getVM();
    for (unsigned long h = 0; h < d; h++)
    {
        DcmTag errortag;
        errortag = errorElements->get(h);
        delete idents -> remove(errortag);
    }
#endif

    /* if there was more than 1 error, override the error comment */
    if (invalidMatchingKeyCount>1)
    {
        errorComment->putString("Syntax error in 1 or more matching keys");
    }

    /* return OFTrue or OFFalse depending on the number of invalid matching attributes */
    return (invalidMatchingKeyCount == 0);
}

OFBool WorklistDB::checkWklist(DcmDataset* dsetfile, const char* fname)
    /*
     * This function checks all elements in the data set information which is contained in the worklist
     * file that is currently being processed if these elements refer to attributes which are supported
     * return attributes. If this is the case, the function will return OFTrue, if this is not the case
     * the function will return OFFalse.
     *
     * Parameters:
     *   dsetfile - [in] The data set information which is contained in the current worklist file.
     *   fname    - [in] Path and filename of the currently processed worklist file .
     */
{
    /* initialize return value */
    OFBool ok = OFFalse;

    /* search for the Scheduled Procedure Step Sequence data */
    /* element in the worklist file's data set information */
    DcmStack stk;
    dsetfile->search(DCM_ScheduledProcedureStepSequence, stk);
    DcmObject *obj = stk.top();

    /* If there is no such data element, we need to return an error; else, do something*/
    if (obj == NULL)
    {
        ok = OFFalse;
        CERR << "WLDB ERROR: missing ScheduledProcedureStepSequence in worklist database file: "
             << fname << endl;
    }
    else
    {
        /* in case there is such a data element, we need to determine if all elements in the */
        /* data set information refer to attributes which are supported return attributes */

        /* determine the number of data elements in the data set */
        unsigned long cdset = dsetfile->card();

        /* start a loop and go through all these elements */
        for (unsigned long i = 0; i < cdset ; i++)
        {
            /* determine the current element */
            DcmElement *element = dsetfile->getElement(i);

            /* determine the current element's length */
            Uint32 length = element->getLength();

            /* determine the current element's tag */
            DcmTag tag(element->getTag());

            /* figure out if the current element refers to an attribute which is a supported return */
            /* attribute. Use level 0 for this check. If that is the case, set ok to OFTrue and go */
            /* ahead. If that is not the case, set ok to OFFalse and terminate the loop */
            if (isaSupportedReturnKey(tag.getXTag(), 0))
                ok = OFTrue;
            else
            {
                ok = OFFalse;
                CERR << "WLDB ERROR: unsupported attribute: "
                     << tag << " " << tag.getTagName() << endl;
                CERR << "            in worklist database file: " << fname << endl;
                break;
            }

            /* if the current element is a sequence */
            if (element->ident() == EVR_SQ)
            {
                /* if its length equals 0, we dont need to scrutinize this element any further */
                if (length == 0)
                    ok = OFTrue;
                else
                {
                    /* if the sequence's length does not equal 0, i.e. the sequence */
                    /* actually contains items, we need to check these items. */

                    /* consider this element as a sequence of items */
                    DcmSequenceOfItems *sequits = (DcmSequenceOfItems*)element;

                    /* get the first item */
                    DcmItem *item = sequits -> getItem(0);

                    /* get its cardinality */
                    unsigned long cit = item -> card();

                    /* start a loop and go through all elements of this item */
                    for (unsigned long k = 0; k < cit; k++)
                    {
                        /* get current element */
                        DcmElement *elementseq = item -> getElement(k);

                        /* determine current element's tag */
                        DcmTag tagseq(elementseq->getTag());

                        /* figure out if the current element refers to an attribute which is a supported return */
                        /* attribute. Use level 1 for this check. If that is the case, set ok to OFTrue and go */
                        /* ahead. If that is not the case, set ok to OFFalse and terminate the loop */
                        if (isaSupportedReturnKey(tagseq.getXTag(), 1))
                            ok = OFTrue;
                        else
                        {
                            ok = OFFalse;
                            CERR << "WLDB ERROR: unsupported attribute: "
                                 << tag << " " << tag.getTagName() << endl;
                            CERR << "            in worklist database file: " << fname << endl;
                            break;
                        }
                    }
                }
            }
        }
    }

    /* return return value */
    return ok;
}

static OFBool
isaWorklistFileName(const char* fname)
    /*
     * This function checks if the given filename refers to a worklist file, i.e. has an
     * extension of ".wl". If it does, the function will return OFTrue, else it will
     * return OFFalse. If a NULL pointer is passed, the function will return OFFalse.
     *
     * Parameters:
     *   fname - [in] Contains the search mask.
     */
{
    /* if no filename is specified, return OFFalse */
    if (fname == NULL)
        return OFFalse;

    /* look for an '.' in the filename */
    char* s = (char *)strrchr(fname, '.');

    /* if there was no '.' return OFFalse */
    if (s == NULL)
        return OFFalse;

    /* if the extension is ".wl" return OFTrue */
    if (strcmp(s, ".wl") == 0)
        return OFTrue;

    /* return OFFalse in all other cases */
    return OFFalse;
}

WDBStatus WorklistDB::startFindRequest(DcmDataset& findRequestIdentifiers)
    /*
     * This function mainly goes through all worklist files and determines those records
     * which match the search mask that was passed. All those records will be written to the
     * member variable objlist. (This variable will be cleared at the beginning of this
     * function so that in the end only those records are contained in this variable.)
     *
     * Parameters:
     *   findRequestIdentifiers - [in] Contains the search mask.
     */
{
    WDBStatus status = WDB_Success;
    OFBool match = OFFalse;
    OFBool *seqmatch = NULL;

    /* initialize offending elements, error elements and error comment */
    delete offendingElements;
    delete errorElements;
    delete errorComment;
    offendingElements = new DcmAttributeTag(DCM_OffendingElement, 0);
    errorElements = new DcmAttributeTag(DCM_OffendingElement, 0);
    errorComment = new DcmLongString(DCM_ErrorComment, 0);

    /* initialize member variable identifiers; this variable will contain the search mask */
    this->clearDataset(identifiers);
    delete identifiers;
    identifiers = new DcmDataset(findRequestIdentifiers);

    /* remove group length and padding elements from the search mask */
    identifiers->computeGroupLengthAndPadding(EGL_withoutGL, EPD_withoutPadding);

    /* remove all elements from member variable objlist; this variable */
    /* will in the end contain all records that match the search mask */
    this->clearObjectList();

    /* this member variable indicates if we encountered an unsupported */
    /* optional key in the search mask; initialize it with false. */
    foundUnsupportedOptionalKey = OFFalse;

    /* scrutinize the search mask */
    if (!this->checkIdentifiers(identifiers))
    {
        /* in case we encountered an error in the search */
        /* mask, we may have to return to the caller */
        if (failInvalidQuery)
        {
            return WDB_Failed_IdentifierDoesNotMatchSOPClass;
        }
    }

    /* dump some information if required */
    if (verbose)
    {
        COUT << "=============================" << endl
             << "Worklist Database Files" << endl;
    }

    /* assign the pointer to a directory path which is con- */
    /* tained in the member variable to a local variable */
    const char * i1dname = dirpath;

    /* if there is no directory path, return */
    if (i1dname == NULL)
        return WDB_Refused_OutOfResorces;

    /* set a read lock on the LOCKFILE in the given directory */
    this -> set_readlock(i1dname);

#ifdef HAVE__FINDFIRST
    struct _finddata_t fileData;
    long hFile;
    OFString fname(i1dname);

    /* determine if the given directory path actually refers to a directory */
    if (((hFile = _findfirst((char *)fname.c_str(), &fileData)) != -1L) &&
        (fileData.attrib & _A_SUBDIR))
    {
        /* if this is the case, we want to read all directory entries in this directory */

        /* first close the search handle hFile again */
        _findclose(hFile);

        /* create a search string */
        OFString searchname;
        searchname = fname + PATH_SEPARATOR + '*';
        int ret = 0;

        /* start a loop; in each iteration another directory entry is determined */
        for (hFile = _findfirst((char *)searchname.c_str(), &fileData);
            hFile != -1L && ret == 0; ret = _findnext(hFile, &fileData))
        {
            /* if the current entry refers to a worklist file, do something */
            if (strcmp(fileData.name, ".")!=0 &&
                strcmp(fileData.name, "..")!=0 &&
                isaWorklistFileName(fileData.name))
            {
                /* create a string that contains path and filename of the current worklist file */
                OFString subname(fname);
                subname += PATH_SEPARATOR;
                subname += fileData.name;
                char file1[1024];
                strcpy(file1, subname.c_str());
#else
    /* determine if the given directory path actually refers to a directory */
    dirp = opendir(i1dname);
    if (dirp != NULL)
    {
        /* if this is the case, we want to read all directory entries in this directory */

        /* start a loop; in each iteration another directory entry is determined */
        struct dirent *dp;
        for (dp = readdir(dirp); dp!=NULL; dp=readdir(dirp))
        {
            /* if the current entry refers to a worklist file, do something */
            if (isaWorklistFileName(dp->d_name))
            {
                /* create a string that contains path and filename of the current worklist file */
                char file1[1024];
                strcpy(file1, i1dname);
                strcat(file1, "/");
                strcat(file1,dp -> d_name);
#endif

                /* dump current file specification (path and filename) */
                COUT << file1 << endl;

                /* associate file with a stream, open in read mode */
                DcmFileStream i1Stream(file1, DCM_ReadMode);

                /* check if stream could be opened */
                if (i1Stream.Fail())
                {
                    CERR << "WLDB: cannot open worklist file: " << file1 << endl;
                    status =  WDB_Refused_OutOfResorces;
                }

                /* read information from worklist file. After the call to DcmFileFormat::read(...) the in-*/
                /* formation which is encapsulated in the file will be available through the DcmFileFormat */
                /* object. In detail, it will be available through calls to DcmFileFormat::getMetaInfo() */
                /* (for meta header information) and DcmFileFormat::getDataset() (for data set information). */
                DcmFileFormat *fileform = new DcmFileFormat();
                fileform->transferInit();
                fileform->read(i1Stream, EXS_Unknown, EGL_withoutGL);
                fileform->transferEnd();

                /* figure out if an error occured while the file was read */
                if (fileform->error() != EC_Normal)
                {
                    CERR << "WLDB: error: " << fileform->error().text() << ": reading worklist file: " << file1 << endl;
                    errorComment->putString("Error in SCP Worklist Database");
                    status = WDB_Failed_UnableToProcess;
                    delete fileform;
                    break;
                }

                /* determine the data set information which is contained in the current worklist file */
                DcmDataset *dsetfile = fileform->getDataset();
                if (dsetfile == NULL)
                {
                    CERR << "WLDB: error: empty worklist file: " << file1 << endl;
                    errorComment->putString("Error in SCP Worklist Database");
                    status = WDB_Failed_UnableToProcess;
                    delete fileform;
                    break;
                }

                /* check if there is a data element in the data set information of the currently */
                /* processed worklist file which does not refer to a supported return attribute. */
                /* If this is the case, we want to return this error */
                if (checkWklist(dsetfile, file1) == OFFalse)
                {
                    delete fileform;
                    errorComment->putString("Error in SCP Worklist Database");
                    status = WDB_Failed_UnableToProcess;
                    break;
                }

                /* if we get to this point we need to go through all data elements */
                /* in the search mask and figure out if they match the corresponding */
                /* elements in the data set of the currently processed worklist file. */
                /* If that is the case, we want to add the corresponding worklist */
                /* record (data set) to the member variable objlist. (This variable */
                /* will in the end contain all records that match the search mask.) */
                /* Note that in the following loop the variable 'match' is used to */
                /* indicate if two corresponding element's match or do not match. */

                /* determine the number of elements in the search mask */
                unsigned long cdset = identifiers->card();

                /* start a loop; go through all the elements in the search mask */
                for (unsigned long i = 0; i < cdset ; ++i)
                {
                    const char *val1;
                    char *val2;

                    /* get the current element of the search mask */
                    DcmElement *element = identifiers->getElement(i);

                    /* determine the current element's length */
                    Uint32 length = element->getLength();

                    /* determine the current element's tag */
                    DcmTagKey tag(element->getTag());

                    /* if the current element is NOT a sequence */
                    if (element->ident() != EVR_SQ)
                    {
                        /* and if it is NOT of data type (VR) UL */
                        if (element->ident() != EVR_UL)
                        {
                            /* if we are looking at a patient's name or a patient's ID */
                            if ((tag == DCM_PatientsName)||(tag == DCM_PatientID))
                            {
                                /* if the length of the search mask's element equals 0, it is considered to be a match */
                                if (length == 0)
                                    match = OFTrue;
                                /* if the length of the search mask's element does not equal 0, we actually */
                                /* have to compare two values to figure out if this is still a match. */
                                else
                                {
                                    /* determine the two corresponding values */
                                    val2 = getStringValue(element);
                                    val1 = findStringValue(dsetfile, tag);

                                    /* and figure out if they match */
                                    if (matchPattern(val1, val2))
                                    {
                                        match = OFTrue;
                                    }
                                    else
                                    {
                                        match = OFFalse;
                                        break;
                                    }
                                }
                            }
                            /* if we are not looking at a patient's name or a patient's ID */
                            else
                            {
                                /* this situation is only considered to be a match if */
                                /* the length of the search mask's element equals 0*/
                                if (length == 0)
                                {
                                    match = OFTrue;
                                }
                            }
                        }
                        /* if the current element is of data type UL */
                        else
                        {
                            /* this situation is only considered to be a match if */
                            /* the length of the search mask's element equals 0*/
                            if (length == 0)
                            {
                                match = OFTrue;
                            }
                        }
                    }
                    /* if the current element IS a sequence */
                    else
                    {
                        /* if the length of the search mask's element equals 0, it is considered to be a match */
                        if (length == 0)
                            match = OFTrue;
                        else
                        {
                            /* if the current element is a sequence and the search pattern's length does not */
                            /* equal 0 we need to consider all the elements of this sequence element. */

                            /* consider the current element as a sequence of items */
                            DcmSequenceOfItems *sequits = (DcmSequenceOfItems*)element;

                            /* get the first item */
                            DcmItem *item = sequits -> getItem(0);

                            /* get the first item's length */
                            Uint32 lengthofit = item->getLength();

                            /* if the first item's length equals 0, it is considered to be a match */
                            if (lengthofit == 0)
                                match = OFTrue;
                            else
                            {
                                /* get the first item's cardinality */
                                unsigned long cit = item -> card();

                                /* search for the same element (the sequence element) */
                                /* in the data set of the worklist file */
                                DcmStack stk1;
                                dsetfile->search(tag, stk1, ESM_fromHere, OFFalse);
                                DcmObject *obj1 = stk1.top();
                                DcmSequenceOfItems *sequits1 = (DcmSequenceOfItems*)obj1;

                                /* get the sequence element's cardinality */
                                unsigned long csequ1 = 0;
                                if (sequits1 != NULL)
                                    csequ1 = sequits1->card();

                                /* create an array of boolean values, one array */
                                /* field for each of the sequence's items */
                                seqmatch = new OFBool[csequ1];

                                /* go through all the sequence element's items (this */
                                /* is the sequence element whithin the worklist file) */
                                for (unsigned long j = 0; j < csequ1; ++j)
                                {
                                    /* get the current item and consider it as a data set */
                                    DcmItem *item1 = sequits1 -> getItem(j);
                                    DcmDataset *dset1 = (DcmDataset*)item1;

                                    /* go through all the elements of the search mask sequence element's first item */
                                    for (unsigned long k = 0; k < cit; ++k)
                                    {
                                        /* get the current element of the search mask sequence element's first item */
                                        DcmElement *elementseq = item -> getElement(k);

                                        /* if the current element is NOT of data type (VR) UL */
                                        if (elementseq->ident()!=EVR_UL)
                                        {
                                            /* get the current element's length */
                                            Uint32 length1 = elementseq -> getLength();

                                            /* get the current element's tag */
                                            DcmTagKey tagseq(elementseq->getTag());

                                            /* if we encountered a certain tag (see below) */
                                            if ((tagseq == DCM_Modality) ||
                                                (tagseq == DCM_ScheduledPerformingPhysiciansName) ||
                                                (tagseq == DCM_ScheduledProcedureStepLocation) ||
                                                (tagseq == DCM_CommentsOnTheScheduledProcedureStep))
                                            {
                                                /* and the length of the (search mask's) current element */
                                                /* equals 0, this is considered to be a match */
                                                if (length1 == 0)
                                                    seqmatch[j] = OFTrue;
                                                /* if the length of the (search mask's) element does not equal 0, we actually */
                                                /* have to compare two values to figure out if this is still a match. */
                                                else
                                                {
                                                    /* determine the two corresponding values */
                                                    val2 = getStringValue(elementseq);
                                                    val1 = findStringValue(dset1, tagseq);

                                                    /* and figure out if they match */
                                                    if (matchPattern(val1, val2))
                                                    {
                                                        seqmatch[j] = OFTrue;
                                                    }
                                                    else
                                                    {
                                                        seqmatch[j] = OFFalse;

                                                        /* in case they don't match we need to */
                                                        /* terminate the corresponding loop */
                                                        break;
                                                    }
                                                }
                                            }
                                            /* if we encountered a certain different tag (see below) */
                                            else if ((tagseq == DCM_ScheduledStationAETitle) ||
                                                     (tagseq ==DCM_ScheduledStationName))
                                            {
                                                /* and the length of the (search mask's) current element */
                                                /* equals 0, this is considered to be a match */
                                                if (length1 == 0)
                                                    seqmatch[j] = OFTrue;
                                                /* if the length of the (search mask's) element does not equal 0, we */
                                                /* need to do something to figure out if this is still a match. */
                                                else
                                                {
                                                    /* the initial assumption is that there is no match */
                                                    seqmatch[j] = OFFalse;

                                                    /* determine the two values which shall be matched */
                                                    val2 = getStringValue(elementseq);
                                                    val1 = findStringValue(dset1, tagseq);

                                                    /* if the whole thing matches, this is for sure considered a match */
                                                    if (matchPattern(val1, val2))
                                                    {
                                                        seqmatch[j] = OFTrue;
                                                    }
                                                    /* if the entire thing did not match, we need to */
                                                    /* match against any multiple values */
                                                    else
                                                    {
                                                        /* determine the value multiplicity (VM) of the string */
                                                        /* (i.e. how many values (which are seperated by '\') are */
                                                        /* stored whithin the string) */
                                                        unsigned long vm = getVMFromString(val1);

                                                        /* go through all the string's values */
                                                        for (unsigned long v = 0; v < vm; v++)
                                                        {
                                                            /* get the current value */
                                                            char * val3;
                                                            val3 = getFirstValueFromString(val1);

                                                            /* figure out if it matches the pattern */
                                                            if (matchPattern(val3, val2))
                                                            {
                                                                seqmatch[j] = OFTrue;
                                                            }
                                                            else
                                                            {
                                                                seqmatch[j] = OFFalse;
                                                            }

                                                            /* free memory */
                                                            delete[] val3;

                                                            /* if the current value did match the */
                                                            /* pattern, we can terminate the loop */
                                                            if (seqmatch[j])
                                                            {
                                                                break;
                                                            }
                                                        }
                                                    }

                                                    /* if the current search mask's element was not matched */
                                                    /* we can terminate the corresponding loop */
                                                    if (!seqmatch[j])
                                                    {
                                                        break;
                                                    }
                                                }
                                            }
                                            /* if we encountered a certain different tag (see below) */
                                            else if (tagseq == DCM_ScheduledProcedureStepStartTime)
                                            {
                                                /* and the length of the (search mask's) current element */
                                                /* equals 0, this is considered to be a match */
                                                if (length1 == 0)
                                                    seqmatch[j] = OFTrue;
                                                /* if the length of the (search mask's) element does not equal 0, we */
                                                /* need to do something to figure out if this is still a match. */
                                                else
                                                {
                                                    /* get the time values which shall be compared */
                                                    val2 = getStringValue(elementseq);
                                                    val1 = findStringValue(dset1, tagseq);

                                                    /* and compare these two values */
                                                    seqmatch[j] = matchTime(val1, val2);

                                                    /* if the current search mask's element was not matched */
                                                    /* we can terminate the corresponding loop */
                                                    if (!seqmatch[j])
                                                    {
                                                        break;
                                                    }
                                                }
                                            }
                                            /* if we encountered a certain different tag (see below) */
                                            else if (tagseq == DCM_ScheduledProcedureStepStartDate)
                                            {
                                                /* and the length of the (search mask's) current element */
                                                /* equals 0, this is considered to be a match */
                                                if (length1 == 0)
                                                    seqmatch[j] = OFTrue;
                                                /* if the length of the (search mask's) element does not equal 0, we */
                                                /* need to do something to figure out if this is still a match. */
                                                else
                                                {
                                                    /* get the date values which shall be compared */
                                                    val2 = getStringValue(elementseq);
                                                    val1 = findStringValue(dset1, tagseq);

                                                    /* and compare these two values */
                                                    seqmatch[j] = matchDate(val1, val2);

                                                    /* if the current search mask's element was not matched */
                                                    /* we can terminate the corresponding loop */
                                                    if (!seqmatch[j])
                                                    {
                                                        break;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }

                                /* go through all array fields of the result array */
                                for (int z = (int)(csequ1-1); z >= 0; z--)
                                {
                                    /* if the current array field shows a value of OFTrue, set 'match' to OFTrue */
                                    if (seqmatch[z] == OFTrue)
                                    {
                                        match = OFTrue;
                                    }
                                    /* else delete the corresponding sequence item from */
                                    /* the sequence element of the worklist file */
                                    else
                                    {
                                        DcmItem *anItem = sequits1 -> remove(z);
                                        delete anItem;
                                    }
                                }

                                /* delete the result array */
                                delete[] seqmatch;

                                /* if there no item in the sequence was matched, we will set 'match' to OFFalse */
                                /* and terminate the loop that goes through all elements in the search mask */
                                if (sequits1 && (sequits1->card() == 0))
                                {
                                    match = OFFalse;
                                    break;
                                }
                            }
                        }
                    }
                }

                /* if the search pattern and the data set from the current worklist file match, */
                /* we need to insert the current DcmFileFormat object (which contains the data */
                /* set that is considered to be a match) into the member variable objlist. Plus, */
                /* we need to figure out a corresponding result value. */
                if (match)
                {
                    objlist->append(fileform);
                    if (foundUnsupportedOptionalKey)
                        status = WDB_Pending_Warning;
                    else
                        status = WDB_Pending;
                }
                /* else we can just delete the current DcmFileFormat object */
                else
                    delete fileform;
            }
        }
#ifdef HAVE__FINDFIRST
        /* close the search handle hFile */
        _findclose(hFile);
#else
        /* perform the opposite to opendir(...) which was called above */
        closedir(dirp);
#endif
    }
    /* if it does not refer to a directory, remember this as an error */
    else
        status = WDB_Refused_OutOfResorces;

    /* release the read lock */
    this -> release_readlock();

    /* set pointer to the first element in objlist */
    objlist->seek(ELP_first);

    /* return result */
    return status;
}





// ********************************


DcmDataset *WorklistDB::nextFindResponse(WDBStatus &rStatus)
    /*
     * This function assumes that function startFindRequest(...) was called earlier and that that
     * function has determined (from the set of worklist files) all those records which match the
     * given search mask. In detail, it is expected that the above mentioned function has written
     * all those records to the member variable objlist. On each call, this function will return
     * one of those records as a (newly created) DcmDataset object. The DcmFileFormat object which
     * is contained in objlist and which refers to the returned object will be removed from objlist.
     *
     * Parameters:
     *   rStatus - [out] .
     */
{
    /* create data set result variable */
    DcmDataset *dsetfile2 = new DcmDataset();

    /* if objlist is empty (i.e. has cardinality 0), */
    /* there are no more records that can be returned */
    if (objlist->card() == 0)
    {
        rStatus = WDB_Success;
        return dsetfile2;
    }

    /* if we get to this point, there is still at least one record (data set) in objlist that has to be returned. */

    /* create some temporary variables */
    WDBStatus status = WDB_Failed_UnableToProcess;
    DcmElement *newElement = NULL;
    DcmItem *newItem = NULL;

    /* determine the (worklist file) data set whose values shall be returned */
    DcmElement *el = (DcmElement*)objlist->seek_to(0);
    DcmFileFormat *theFile = (DcmFileFormat*)el;
    DcmDataset *set = theFile->getDataset();

    /*the variable 'identifiers' containes the search mask. Determine it's cardinality */
    unsigned long int cids = identifiers->card();

    /* go through all elements of the search mask */
    for (unsigned long int j = 0; j < cids; j++)
    {
        /* get the current (search mask) element */
        DcmElement *element = identifiers->getElement(j);

        /* get the current (search mask) element's tag */
        DcmTag tag(element->getTag());

        /* if the current (search mask) element is NOT a sequence */
        if (element->ident() != EVR_SQ)
        {
            /* if the current (search mask) element is a supported return attribute, we */
            /* need to return it (insert it into the result data set variable) */
            if (isaSupportedReturnKey(tag.getXTag(), 0))
            {
                /* in case the current (search mask) element is of VR (data type) US, like this */
                if (element->ident() == EVR_US)
                {
                  newDicomElement(newElement, tag);
                  Uint16 *USval = findUSValue(set, tag);
                  if (USval) newElement -> putUint16(*USval,0);
                  dsetfile2 -> insert(newElement, OFTrue /*replaceOld*/);
                }
                /* in all other cases like this */
                else
                {
                  newDicomElement(newElement, tag);
                  char * val1 = findStringValue(set, tag);
                  if (val1) newElement -> putString(val1);
                  dsetfile2 -> insert(newElement, OFTrue /*replaceOld*/);
                }
            }
            /* if the current (search mask) element is not a supported return attribute, */
            /* insert an empty element into the result data set variable */
            else
            {
                newDicomElement(newElement, tag);
                dsetfile2 -> insert(newElement, OFTrue /*replaceOld*/);
            }
        }
        /* if the current (search mask) element IS a sequence */
        else
        {
            /* consider this (search mask) element as a sequence of items */
            DcmSequenceOfItems *sequits = (DcmSequenceOfItems*)element;

            /* create a new (result) DcmElement object based on the (search mask's) sequence element's tag */
            newDicomElement(newElement, tag, 0);

            /* consider this new (result) DcmElement object as a sequence of items */
            DcmSequenceOfItems *newsequits = (DcmSequenceOfItems *)newElement;

            /* insert this (result) DcmSequenceOfItems object into the result data set variable */
            dsetfile2 -> insert(newsequits, OFTrue /*replaceOld*/);

            /* search for the sequence of items element in */
            /* the (worklist-file) data set object */
            DcmStack stk1;
            set->search(tag.getXTag(), stk1, ESM_fromHere, OFFalse);
            DcmObject *obj1 = stk1.top();
            DcmSequenceOfItems *sequits1 = (DcmSequenceOfItems*)obj1;

            /* get the (worklist-file's data set's) sequence element's cardinality */
            unsigned long csequ1 = 0;
            if (sequits1)
                csequ1 = sequits1->card();

            /* go through all the (worklist-file's data set's) sequence element's items */
            for (unsigned long int z = 0; z < csequ1; z++)
            {
                /* get the current (worklist-file's, data set's, sequence element's) item */
                DcmItem *item1 = sequits1->getItem(z);

                /* create a new DcmItem object */
                newItem = new DcmItem();

                /* insert this object into the above created (result) DcmSequenceOfItems object */
                newsequits -> insert(newItem);

                /* initialize an indicator variable that indicates if the search */
                /* mask's sequence of items element is actually empty */
                OFBool emptyIdentifiersItem = OFFalse;
                if (sequits->card() == 0)
                {
                    emptyIdentifiersItem = OFTrue;
                }
                else
                {
                    DcmItem *item = sequits->getItem(0);
                    if ((item == NULL) || (item->card() == 0))
                    {
                        emptyIdentifiersItem = OFTrue;
                    }
                }

                /* if it actually IS empty */
                if (emptyIdentifiersItem)
                {
                    /* go through all elements of the (worklist-file's, data set's, sequence element's) current element */
                    unsigned long cit1 = item1->card();
                    for (unsigned long int w = 0; w < cit1; w++)
                    {
                        /* get the current element */
                        DcmElement *elementseq1 = item1 -> getElement(w);

                        /* get the current element's tag */
                        DcmTag tagseq1(elementseq1->getTag());

                        /* if the current element is a supported return attribute, we */
                        /* need to return it (insert it into the result data set variable) */
                        if (isaSupportedReturnKey(tagseq1.getXTag(), 1))
                        {
                            newDicomElement(newElement, tagseq1);
                            char *val1 = getStringValue(elementseq1);
                            newElement -> putString(val1);
                            newItem -> insert(newElement, OFTrue /*replaceOld*/);
                        }
                        /* if the current element is not a supported return attribute, */
                        /* insert an empty element into the data set variable */
                        else
                        {
                            newDicomElement(newElement, tagseq1);
                            newItem -> insert(newElement, OFTrue /*replaceOld*/);
                        }
                    }
                }
                /* if it is NOT empty */
                else
                {
                    /* get the first item of the search mask's sequence of items element */
                    DcmItem *item = sequits->getItem(0);

                    /* get its cardinality */
                    unsigned long cit = item->card();

                    /* go through all elements of the first item of the search mask's sequence of items element */
                    for (unsigned long k = 0; k < cit; ++k)
                    {
                        /* get the first element */
                        DcmElement *elementseq = item -> getElement(k);

                        /* get the first element's tag */
                        DcmTag tagseq(elementseq->getTag());

                        /* if the current element is a supported return attribute, we */
                        /* need to return it (insert it into the result data set variable) */
                        if (isaSupportedReturnKey(tagseq.getXTag(), 1))
                        {
                            newDicomElement(newElement, tagseq);
                            char *val1 = findStringValue(item1, tagseq);
                            newElement -> putString(val1);
                            newItem -> insert(newElement, OFTrue /*replaceOld*/);
                        }
                        /* if the current element is not a supported return attribute, */
                        /* insert an empty element into the data set variable */
                        else
                        {
                            newDicomElement(newElement, tagseq);
                            newItem -> insert(newElement, OFTrue /*replaceOld*/);
                        }
                    }
                }
            }
        }
    }

    /* determine a result value for the out parameter */
    if (objlist->card() > 0)
    {
        if (foundUnsupportedOptionalKey)
            status = WDB_Pending_Warning;
        else
            status = WDB_Pending;

        /* remove the current element from objlist */
        objlist->remove();
    }

    /* free memory */
    delete theFile;

    /* set status correspondingly */
    rStatus = status;

    /* return the data set which was created */
    return dsetfile2;
}




WDBStatus WorklistDB::cancelFindRequest()
{
   this -> clearObjectList();

   return WDB_Cancel;
}

const DcmAttributeTag* WorklistDB::getOffendingElements()
{
    if (offendingElements->getLength() == 0)
        return NULL;
    else
        return offendingElements;
}

const DcmLongString* WorklistDB::getErrorComments()
{
    if (errorComment->getLength() == 0)
        return NULL;
    else
        return errorComment;
}


/*
** CVS Log
** $Log: wrklstdb.cc,v $
** Revision 1.36  2001/11/28 13:43:07  joergr
** Check return value of DcmItem::insert() statements where appropriate to
** avoid memory leaks when insert procedure fails.
**
** Revision 1.35  2001/11/06 14:20:31  wilkens
** Added more comments.
**
** Revision 1.34  2001/09/26 15:57:09  meichel
** Adapted wlistctn to class OFCondition
**
** Revision 1.33  2001/06/01 15:51:46  meichel
** Updated copyright header
**
** Revision 1.32  2000/12/15 13:25:35  meichel
** Declared qsort() and signal() callback functions as extern "C", avoids
**   warnings on Sun C++ 5.x compiler.
**
** Revision 1.31  2000/03/08 16:37:48  meichel
** Updated copyright header.
**
** Revision 1.30  2000/03/03 14:18:24  meichel
** Implemented library support for redirecting error messages into memory
**   instead of printing them to stdout/stderr for GUI applications.
**
** Revision 1.29  2000/02/23 15:14:01  meichel
** Corrected macro for Borland C++ Builder 4 workaround.
**
** Revision 1.28  2000/02/02 14:51:41  joergr
** Replaced 'delete' statements by 'delete[]' for objects created with 'new[]'.
**
** Revision 1.27  2000/02/01 11:01:01  meichel
** Avoiding to include <stdlib.h> as extern "C" on Borland C++ Builder 4,
**   workaround for bug in compiler header files.
**
** Revision 1.26  1999/11/12 16:51:27  meichel
** Corrected file locking code that did not work correctly under Win95/98.
**
** Revision 1.25  1999/04/30 16:37:35  meichel
** Renamed all flock calls to dcmtk_flock to avoid name clash between flock()
** emulation based on fcntl() and a constructor for struct flock.
**
** Revision 1.24  1999/04/26 12:59:33  meichel
** Ported Worklist CTN to Win32 environment.
**
** Revision 1.23  1999/03/22 09:56:44  meichel
** Reworked data dictionary based on the 1998 DICOM edition and the latest
**   supplement versions. Corrected dcmtk applications for minor changes
**   in attribute name constants.
**
** Revision 1.22  1998/09/03 10:09:14  meichel
** The Worklist SCP now supports all Type 1 and 2 attributes
**   defined in the Modality Worklist Service Class as return keys.
**
** Revision 1.21  1998/07/24 13:29:13  meichel
** changed WorklistDB::nextFindResponse() parameters
**   in order to avoid the use of an implicit assignment
**   operator for class DcmDataset.
**
** Revision 1.20  1998/02/06 15:07:46  meichel
** Removed many minor problems (name clashes, unreached code)
**   reported by Sun CC4 with "+w" or Sun CC2.
**
** Revision 1.19  1998/01/14 15:43:23  hewett
** Added support for the VRs UT (Unlimited Text) and VS (Virtual String).
**
** Revision 1.18  1997/09/18 14:41:30  meichel
** Some systems, e.g. NeXTStep, need the third argument
**   for fcntl calls to be casted to int. Other systems,
**   e.g. OSF1-Alpha, won't accept this because int and struct flock *
**   have different sizes. The workaround used here is to use a typecast to int
**   if sizeof(void *) == sizeof(int) and leave it away otherwise.
**
** Revision 1.17  1997/09/18 08:11:07  meichel
** Many minor type conflicts (e.g. long passed as int) solved.
**
** Revision 1.16  1997/07/21 09:09:01  andreas
** - Replace all boolean types (BOOLEAN, CTNBOOLEAN, DICOM_BOOL, BOOL)
**   with one unique boolean type OFBool.
**
** Revision 1.15  1997/07/07 07:53:29  andreas
** - Changed parameter type DcmTag & to DcmTagKey & in all search functions
**   in DcmItem, DcmSequenceOfItems, DcmDirectoryRecord and DcmObject
**
** Revision 1.14  1997/05/29 09:09:07  hewett
** added explicit cast of third argument for fcntl().  Avoids a GCC compiler
** warning under Nextstep.
**
** Revision 1.13  1997/05/22 13:36:57  hewett
** Modified the test for presence of a data dictionary to use the
** method DcmDataDictionary::isDictionaryLoaded().
**
** Revision 1.12  1997/05/16 08:31:45  andreas
** - Revised handling of GroupLength elements and support of
**   DataSetTrailingPadding elements. The enumeratio E_GrpLenEncoding
**   got additional enumeration values (for a description see dctypes.h).
**   addGroupLength and removeGroupLength methods are replaced by
**   computeGroupLengthAndPadding. To support Padding, the parameters of
**   element and sequence write functions changed.
**
** Revision 1.11  1997/04/18 08:40:41  andreas
** - The put/get-methods for all VRs did not conform to the C++-Standard
**   draft. Some Compilers (e.g. SUN-C++ Compiler, Metroworks
**   CodeWarrier, etc.) create many warnings concerning the hiding of
**   overloaded get methods in all derived classes of DcmElement.
**   So the interface of all value representation classes in the
**   library are changed rapidly, e.g.
**   OFCondition get(Uint16 & value, const unsigned long pos);
**   becomes
**   OFCondition getUint16(Uint16 & value, const unsigned long pos);
**   All (retired) "returntype get(...)" methods are deleted.
**   For more information see dcmdata/include/dcelem.h
**
** Revision 1.10  1997/01/13 15:35:19  hewett
** Fixed crashing bug in the worklist database search routines.  A badly
** formed date or time matching key was not being caught.  Also added
** some preliminary error checking of maching keys prior to database
** search.  Malformed keys will be noted as offending elements and the
** database search routines will return an error.
**
** Revision 1.9  1996/09/27 08:50:15  hewett
** Corrected assignment of const char* to char* (for IBM AIX C Set++).
**
** Revision 1.8  1996/05/30 17:55:22  hewett
** Corrected processing of sequences in query identifiers.  Any and
** every sequence was being treated (and matched) as if it were a
** ScheduledProcedureStepSequence.
**
** Revision 1.7  1996/05/02 14:48:18  hewett
** Unsupported request identifiers as now returned in responses
** with no value.  If unsupported request identifiers are detected
** and a match is found then a status of Pending_Warning is generated
** instead of Pending.
**
** Revision 1.6  1996/05/02 13:42:10  hewett
** Revised error handling.  More accurate error messages when db is bad.
**
** Revision 1.5  1996/05/01 17:34:31  hewett
** Corrected some matching errors.  Removed some memory leaks.
**
** Revision 1.4  1996/04/30 13:57:44  savonge
** - more errors corrected
** - added checking of wklist database
**
** Revision 1.3  1996/04/29 19:15:08  hewett
** Added header comment and conditional inclusion of various .h files.
**
**
*/
